from __future__ import annotations

from typing import TYPE_CHECKING

from attr import attrib, attrs
from marshmallow import Schema, fields

from pytest_bdd.utils import ModelSchemaPostLoadable


@attrs
class AST(ModelSchemaPostLoadable):
    gherkin_document: GherkinDocument = attrib()

    @property
    def registry(self):
        return {**(self.gherkin_document.feature.registry if hasattr(self.gherkin_document, "feature") else {})}


class ASTSchema(Schema):
    gherkin_document = fields.Nested(lambda: GherkinDocumentSchema(), data_key="gherkinDocument")

    build_ast = AST.schema_post_loader()


@attrs
class GherkinDocument(ModelSchemaPostLoadable):
    comments: list[Comment] = attrib()
    uri: str = attrib()

    # Workaround because of allure integration
    if TYPE_CHECKING:  # pragma: no cover
        feature: Feature = attrib(init=False)

    postbuild_attrs = ["feature"]


class GherkinDocumentSchema(Schema):
    comments = fields.Nested(lambda: CommentSchema(many=True))
    feature = fields.Nested(lambda: FeatureSchema(), required=False)
    uri = fields.Str()

    build_gherkin_document = GherkinDocument.schema_post_loader()


@attrs
class Identifiable:
    identifier: str = attrib()

    @property
    def registry(self):
        return {self.identifier: self}


class IdentifiableSchema(Schema):
    identifier = fields.Str(data_key="id")


@attrs
class Tagable:
    tags: list[Tag] = attrib()

    @property
    def registry(self):
        registry = {}
        for tag in self.tags:
            registry.update(tag.registry)
        return registry


class TagableSchema(Schema):
    tags = fields.Nested(lambda: TagSchema(many=True))


@attrs
class Locatable:
    location: Location = attrib()


class LocatableSchema(Schema):
    location = fields.Nested(lambda: LocationSchema())


@attrs
class Keywordable:
    keyword: str = attrib()


class KeywordableSchema(Schema):
    keyword = fields.Str()


@attrs
class Descriptable:
    description: str = attrib()


class DescriptableSchema(Schema):
    description = fields.Str()


@attrs
class Nameable:
    name: str = attrib()


class NameableSchema(Schema):
    name = fields.Str()


@attrs
class Tag(Identifiable, Locatable, Nameable, ModelSchemaPostLoadable):
    ...


class TagSchema(IdentifiableSchema, LocatableSchema, NameableSchema):
    build_tag = Tag.schema_post_loader()


@attrs
class Comment(Locatable, ModelSchemaPostLoadable):
    text: str = attrib()


class CommentSchema(LocatableSchema, Schema):
    text = fields.Str()

    build_comment = Comment.schema_post_loader()


@attrs
class NodeProto(Locatable, Keywordable, Descriptable, Nameable):
    ...


class NodeProtoSchema(LocatableSchema, KeywordableSchema, DescriptableSchema, NameableSchema, Schema):
    ...


@attrs
class NodeContainerProto(NodeProto, Tagable):
    children: list[NodeContainerChild] = attrib()

    @property
    def registry(self):
        registry = {**Tagable.registry.__get__(self)}
        for child in self.children:
            registry.update(child.registry)
        return registry


class NodeContainerProtoSchema(NodeProtoSchema, TagableSchema, Schema):
    children = fields.Nested(lambda: NodeContainerChildSchema(many=True))


@attrs
class NodeStepContainerProto(NodeProto):
    steps: list[Step] = attrib()

    @property
    def registry(self):
        registry = {}
        for step in self.steps:
            registry.update(step.registry)
        return registry


class NodeStepContainerProtoSchema(NodeProtoSchema, Schema):
    steps = fields.Nested(lambda: StepSchema(many=True))


@attrs
class Feature(NodeContainerProto, ModelSchemaPostLoadable):
    language: str = attrib()
    keyword: str = attrib(init=False)

    postbuild_attrs = ["keyword"]


class FeatureSchema(NodeContainerProtoSchema, Schema):
    language = fields.Str()
    keyword = fields.Str(required=False)

    build_feature = Feature.schema_post_loader()


@attrs
class Rule(NodeContainerProto, Identifiable, ModelSchemaPostLoadable):
    @property
    def registry(self):
        return {
            **NodeContainerProto.registry.__get__(self),
            **Identifiable.registry.__get__(self),
        }


class RuleSchema(NodeContainerProtoSchema, IdentifiableSchema, Schema):
    build_rule = Rule.schema_post_loader()


@attrs
class Background(NodeStepContainerProto, Identifiable, ModelSchemaPostLoadable):
    @property
    def registry(self):
        return {
            **NodeStepContainerProto.registry.__get__(self),
            **Identifiable.registry.__get__(self),
        }


class BackgroundSchema(NodeStepContainerProtoSchema, IdentifiableSchema, Schema):
    build_background = Background.schema_post_loader()


@attrs
class Scenario(NodeStepContainerProto, Identifiable, Tagable, ModelSchemaPostLoadable):
    examples: list[Example] = attrib()

    @property
    def registry(self):
        registry = {
            **NodeStepContainerProto.registry.__get__(self),
            **Identifiable.registry.__get__(self),
            **Tagable.registry.__get__(self),
        }
        for example in self.examples:
            registry.update(example.registry)
        return registry


class ScenarioSchema(NodeStepContainerProtoSchema, IdentifiableSchema, TagableSchema, Schema):
    examples = fields.Nested(lambda: ExampleSchema(many=True))

    build_scenario = Scenario.schema_post_loader()


@attrs
class NodeContainerChild(ModelSchemaPostLoadable):

    # Workaround because of allure integration
    if TYPE_CHECKING:  # pragma: no cover
        background: Background = attrib(init=False)
        scenario: Scenario = attrib(init=False)
        rule: Rule = attrib(init=False)

    postbuild_attrs = ["background", "scenario", "rule"]

    @property
    def registry(self):
        registry = {
            **(self.background.registry if hasattr(self, "background") else {}),
            **(self.scenario.registry if hasattr(self, "scenario") else {}),
            **(self.rule.registry if hasattr(self, "rule") else {}),
        }
        return registry


class NodeContainerChildSchema(Schema):
    background = fields.Nested(lambda: BackgroundSchema(), required=False)
    scenario = fields.Nested(lambda: ScenarioSchema(), required=False)
    rule = fields.Nested(lambda: RuleSchema(), required=False)

    build_node_container_child = NodeContainerChild.schema_post_loader()


@attrs
class Location(ModelSchemaPostLoadable):
    column: int = attrib()
    line: int = attrib()


class LocationSchema(Schema):
    column = fields.Int()
    line = fields.Int()

    build_loc = Location.schema_post_loader()


@attrs
class DocString(Locatable, ModelSchemaPostLoadable):
    content: str = attrib()
    delimiter: str = attrib()

    # Workaround because of allure integration
    if TYPE_CHECKING:  # pragma: no cover
        media_type: str = attrib(init=False)

    postbuild_attrs = ["media_type"]


class DocStringSchema(LocatableSchema, Schema):
    content = fields.Str()
    delimiter = fields.Str()
    media_type = fields.Str(required=False, data_key="mediaType")

    build_docstring = DocString.schema_post_loader()


@attrs
class Step(Identifiable, Keywordable, Locatable, ModelSchemaPostLoadable):
    text: str = attrib()
    keyword_type: str = attrib()

    # Workaround because of allure integration
    if TYPE_CHECKING:  # pragma: no cover
        data_table: DataTable = attrib(init=False)
        doc_string: DocString = attrib(init=False)

    postbuild_attrs = ["data_table", "doc_string"]

    @property
    def registry(self):
        registry = {
            **Identifiable.registry.__get__(self),
            **(self.data_table.registry if hasattr(self, "data_table") else {}),
        }
        return registry


class StepSchema(IdentifiableSchema, KeywordableSchema, LocatableSchema, Schema):
    text = fields.Str()
    keyword_type = fields.Str(data_key="keywordType", required=False, default="Unspecified")
    data_table = fields.Nested(lambda: DataTableSchema(), data_key="dataTable", required=False)
    doc_string = fields.Nested(lambda: DocStringSchema(), data_key="docString", required=False)

    build_step = Step.schema_post_loader()


@attrs
class Example(Descriptable, Identifiable, Keywordable, Locatable, Nameable, Tagable, ModelSchemaPostLoadable):
    table_body: list[TableRow] = attrib()

    # Workaround because of allure integration
    if TYPE_CHECKING:  # pragma: no cover
        table_header: ExampleTableHeader = attrib(init=False)

    postbuild_attrs = ["table_header"]

    @property
    def registry(self):
        registry = {
            **Identifiable.registry.__get__(self),
            **(self.table_header.registry if hasattr(self, "table_header") else {}),
        }
        for table_row in self.table_body:
            registry.update(table_row.registry)
        return registry


class ExampleSchema(
    DescriptableSchema, IdentifiableSchema, KeywordableSchema, LocatableSchema, NameableSchema, TagableSchema, Schema
):
    table_body = fields.Nested(lambda: TableRowSchema(many=True), data_key="tableBody")
    table_header = fields.Nested(lambda: TableHeaderSchema(), data_key="tableHeader", required=False)

    build_example = Example.schema_post_loader()


@attrs
class TableRow(Identifiable, Locatable, ModelSchemaPostLoadable):
    cells: list[TableCell] = attrib()


class TableRowSchema(IdentifiableSchema, LocatableSchema, Schema):
    cells = fields.Nested(lambda: TableCellSchema(many=True))

    build_table_row = TableRow.schema_post_loader()


@attrs
class ExampleTableHeader(TableRow):
    ...


class TableHeaderSchema(TableRowSchema, Schema):
    ...


@attrs
class TableCell(Locatable, ModelSchemaPostLoadable):
    value: str = attrib()


class TableCellSchema(LocatableSchema, Schema):
    value = fields.Str()

    build_cell = TableCell.schema_post_loader()


@attrs
class DataTable(Locatable, ModelSchemaPostLoadable):
    rows: list[TableRow] = attrib()

    @property
    def registry(self):
        registry = {}
        for row in self.rows:
            registry.update(row.registry)
        return registry


class DataTableSchema(LocatableSchema, Schema):
    rows = fields.Nested(lambda: TableRowSchema(many=True))

    build_datatable = DataTable.schema_post_loader()



"""
Compatibility module for pytest
"""
from __future__ import annotations

import sys
from operator import ge
from typing import TYPE_CHECKING

from _pytest.config import Config, PytestPluginManager
from _pytest.config.argparsing import Parser
from _pytest.fixtures import FixtureDef, FixtureLookupError, call_fixture_func
from _pytest.main import Session, wrap_session
from _pytest.mark import Mark, MarkDecorator
from _pytest.python import Metafunc
from _pytest.reports import TestReport
from _pytest.runner import CallInfo
from _pytest.terminal import TerminalReporter

from pytest_bdd.packaging import compare_distribution_version

assert all(
    [
        CallInfo,
        call_fixture_func,
        Config,
        FixtureDef,
        FixtureLookupError,
        Mark,
        MarkDecorator,
        Metafunc,
        Parser,
        PytestPluginManager,
        Session,
        TerminalReporter,
        TestReport,
        wrap_session,
    ]
)

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias
assert TypeAlias

# region pytest version dependent imports
if compare_distribution_version("pytest", "7.0", ge):
    if TYPE_CHECKING:  # pragma: no cover
        from pytest import Testdir
else:
    if TYPE_CHECKING:  # pragma: no cover
        from _pytest.pytester import Testdir  # type: ignore[no-redef, attr-defined]

if compare_distribution_version("pytest", "6.2", ge):
    from pytest import FixtureRequest
else:
    from _pytest.fixtures import FixtureRequest
assert FixtureRequest

# endregion

if TYPE_CHECKING:  # pragma: no cover
    from _pytest.nodes import Item as BaseItem
    from _pytest.pytester import RunResult

    assert RunResult
    assert Testdir

    class Item(BaseItem):
        _request: FixtureRequest

else:
    from _pytest.nodes import Item

    assert Item
assert Item



"""pytest-bdd missing test code generation."""
from __future__ import annotations

import itertools
import os.path
from functools import reduce
from operator import lt, methodcaller
from pathlib import Path
from typing import Any, cast

import py
from mako.lookup import TemplateLookup
from pytest import ExitCode

from pytest_bdd.const import STEP_TYPE_BY_NORMALIZED_PREFIX
from pytest_bdd.model import Feature, Scenario, Step
from pytest_bdd.packaging import compare_distribution_version
from pytest_bdd.parser import GherkinParser
from pytest_bdd.steps import StepHandler
from pytest_bdd.typing.pytest import Config, FixtureRequest, Item, Parser, Session, wrap_session
from pytest_bdd.utils import make_python_name

template_lookup = TemplateLookup(directories=[os.path.join(os.path.dirname(__file__), "templates")])


def add_options(parser: Parser) -> None:
    """Add pytest-bdd options."""
    group = parser.getgroup("bdd", "Generation")

    group._addoption(
        "--generate-missing",
        action="store_true",
        dest="generate_missing",
        default=False,
        help="Generate missing bdd test code for given feature files and exit.",
    )

    group._addoption(
        "--feature",
        metavar="FILE_OR_DIR",
        action="append",
        dest="features",
        help="Feature file or directory to generate missing code for. Multiple allowed.",
    )


def cmdline_main(config: Config) -> int | None:
    """Check config option to show missing code."""
    if config.option.generate_missing:
        return show_missing_code(config)
    return None  # Make mypy happy


def generate_code(features: list[Feature], scenarios: list[Scenario], steps: list[Step]) -> str:
    """Generate test code for the given filenames."""
    grouped_steps = group_steps(steps)
    template = template_lookup.get_template("test.py.mak")
    code = template.render(
        features=features,
        scenarios=scenarios,
        steps=grouped_steps,
        make_python_name=make_python_name,
        make_python_docstring=make_python_docstring,
        make_string_literal=make_string_literal,
    )
    return cast(str, code)


def show_missing_code(config: Config) -> int | ExitCode:
    """Wrap pytest session to show missing code."""

    return wrap_session(config, _show_missing_code_main)


def print_missing_code(scenarios: list[Scenario], steps: list[Step]) -> None:
    """Print missing code with TerminalWriter."""
    tw = py.io.TerminalWriter()
    scenario = step = None

    for scenario in scenarios:
        tw.line()
        tw.line(
            f'Scenario "{scenario.name}" is not bound to any test in the feature "{scenario.feature.name}"'
            f" in the file {scenario.feature.filename}:{scenario.line_number}",
            red=True,
        )

    if scenario:
        tw.sep("-", red=True)

    for step in steps:
        tw.line()
        tw.line(
            f"""StepHandler {step.keyword} "{step.name}" is not defined in the scenario "{step.scenario.name}" in the feature"""
            f""" "{step.scenario.feature.name}" in the file"""
            f""" {step.scenario.feature.filename}:{step.line_number}""",
            red=True,
        )

    if step:
        tw.sep("-", red=True)

    tw.line("Please place the code above to the test file(s):")
    tw.line()

    features: list[Feature] = sorted(
        reduce(
            lambda list_, item: cast(list, list_) + [item] if item not in list_ else list_,
            [scenario.feature for scenario in scenarios],
            [],
        ),
        key=lambda feature: feature.uri,  # type: ignore[no-any-return]  # https://github.com/python/typing/issues/760
    )
    code = generate_code(features, scenarios, steps)
    tw.write(code)


def parse_feature_files(paths: list[str], **kwargs: Any) -> tuple[list[Feature], list[Scenario], list[Step]]:
    """Parse feature files of given paths.

    :param paths: `list` of paths (file or dirs)

    :return: `list` of `tuple` in form:
             (`list` of `Feature` objects, `list` of `Scenario` objects, `list` of `StepHandler` objects).
    """
    features = GherkinParser().get_from_paths(list(map(Path, paths)), **kwargs)
    scenarios = list(
        tuple(
            zip(
                *sorted(
                    itertools.chain.from_iterable(
                        ((feature, scenario) for scenario in feature.scenarios) for feature in features
                    ),
                    key=lambda item: (item[0].uri, item[1].id, item[1].name),
                )
            )
        )[1]
    )

    seen = set()
    steps: list[Step] = sorted(
        (
            seen.add(step.name) or step  # type: ignore[func-returns-value]
            for step in itertools.chain.from_iterable(scenario.steps for scenario in scenarios)
            if step.name not in seen
        ),
        key=lambda step: step.name,  # type: ignore[no-any-return]  # https://github.com/python/typing/issues/760
    )
    return features, scenarios, steps


def group_steps(steps: list[Step]) -> list[Step]:
    """Group steps by type."""
    steps = sorted(steps, key=lambda step: STEP_TYPE_BY_NORMALIZED_PREFIX[step.prefix])
    seen_steps = set()
    grouped_steps: list[Step] = []
    for step in itertools.chain.from_iterable(
        sorted(group, key=lambda step: step.name)  # type: ignore[no-any-return]  # https://github.com/python/typing/issues/760
        for _, group in itertools.groupby(
            steps, lambda step: sorted(steps, key=lambda step: STEP_TYPE_BY_NORMALIZED_PREFIX[step.prefix])
        )
    ):
        if step.name not in seen_steps:
            grouped_steps.append(step)
            seen_steps.add(step.name)
    grouped_steps.sort(key=lambda step: step.prefix)  # type: ignore[no-any-return]  # https://github.com/python/typing/issues/760
    return grouped_steps


def _show_missing_code_main(config: Config, session: Session) -> None:
    """Preparing fixture duplicates for output."""
    tw = py.io.TerminalWriter()
    config.hook.pytest_collection(session=session)

    if config.option.features is None:
        tw.line("The --feature parameter is required.", red=True)
        session.exitstatus = 100
        return

    features, scenarios, steps = parse_feature_files(config.option.features)
    scenarios = list(scenarios)

    for item in session.items:

        is_legacy_pytest = compare_distribution_version("pytest", "7.0", lt)

        method_name = "prepare" if is_legacy_pytest else "setup"
        methodcaller(method_name, item)(item.session._setupstate)

        item = cast(Item, item)
        item_request: FixtureRequest = item._request
        scenario: Scenario = item_request.getfixturevalue("scenario")
        feature: Feature = scenario.feature

        for i, s in enumerate(scenarios):
            if s.id == scenario.id and s.uri == scenario.uri:
                scenarios.remove(s)
                break

        previous_step = None
        for step in scenario.steps:
            try:
                item_request.config.hook.pytest_bdd_match_step_definition_to_step(
                    request=item_request, feature=feature, scenario=scenario, step=step, previous_step=previous_step
                )
            except StepHandler.Matcher.MatchNotFoundError:
                pass
            else:
                steps = [non_found_steps for non_found_steps in steps if non_found_steps.name != step.name]
            finally:
                previous_step = step

        item.session._setupstate.teardown_exact(*((item,) if is_legacy_pytest else ()), None)  # type: ignore[call-arg]

    grouped_steps = group_steps(steps)
    print_missing_code(scenarios, grouped_steps)

    if scenarios or steps:
        session.exitstatus = 100


def make_python_docstring(string: str) -> str:
    """Make a python docstring literal out of a given string."""
    return '"""{}."""'.format(string.replace('"""', '\\"\\"\\"'))


def make_string_literal(string: str) -> str:
    """Make python string literal out of a given string."""
    return "'{}'".format(string.replace("'", "\\'"))



from __future__ import annotations

import os

import pytest
from attr import asdict
from pluggy import HookimplMarker

from pytest_bdd.model import Feature, Scenario, Step, Tag
from pytest_bdd.typing.allure import ALLURE_INSTALLED

if ALLURE_INSTALLED:
    from allure_commons import hookimpl
    from allure_commons import plugin_manager as allure_plugin_manager
    from allure_commons._allure import StepContext
    from allure_commons.model2 import Label, Parameter, Status, TestStepResult
    from allure_commons.types import LabelType
    from allure_commons.utils import md5, now, platform_label
    from allure_pytest.listener import AllureListener
else:
    hookimpl = HookimplMarker("allure")


class AllurePytestBDD:
    def __init__(self, allure_logger, allure_cache):
        self.allure_logger = allure_logger
        self._cache = allure_cache

    @classmethod
    def register_if_allure_accessible(cls, config):
        pluginmanager = config.pluginmanager
        allure_accessible = pluginmanager.hasplugin("allure_pytest") and config.option.allure_report_dir
        if allure_accessible:
            allure_plugin_manager.get_plugins()

            listener = next(
                filter(lambda plugin: isinstance(plugin, AllureListener), allure_plugin_manager.get_plugins())
            )

            bdd_listener = cls(listener.allure_logger, listener._cache)
            allure_plugin_manager.register(bdd_listener)
            pluginmanager.register(bdd_listener)

    @hookimpl(hookwrapper=True)
    def report_result(self, result):
        recomposable_features = []

        def decompose_scenario_or_feature(attr, value):
            if isinstance(value, (Tag, Step, Scenario, Feature)):
                if isinstance(value, Feature):
                    recomposable_features.append(value)
                value.decompose()
            return True

        asdict(result, filter=decompose_scenario_or_feature)

        yield

        while recomposable_features:
            recomposable_features.pop().compose()

    @pytest.hookimpl
    def pytest_bdd_before_step_call(self, request, feature, scenario, step, step_func, step_func_args, step_definition):
        """Called before step function is set up."""
        step_definition.func = StepContext(f"{step.keyword} {step.name}", step_func_args)(step_func)

    @pytest.hookimpl
    def pytest_bdd_before_scenario(self, request, feature, scenario):

        scenario_result_uuid = self._cache.get(scenario)
        test_result_uuid = self._cache.get(request.node.nodeid)

        if not scenario_result_uuid:
            scenario_result_uuid = self._cache.push(scenario)

        self.allure_logger.start_step(test_result_uuid, scenario_result_uuid, TestStepResult())

        scenario_result = self.allure_logger.get_test(scenario_result_uuid)
        test_result = self.allure_logger.get_item(test_result_uuid)
        full_name = self.get_full_name(feature, scenario)
        name = self.get_name(request.node, scenario)

        scenario_result.fullName = full_name
        scenario_result.name = name
        scenario_result.start = now()
        scenario_result.historyId = md5(request.node.nodeid)
        test_result.labels.append(Label(name=LabelType.FRAMEWORK, value="pytest-bdd"))
        test_result.labels.append(Label(name=LabelType.LANGUAGE, value=platform_label()))
        test_result.labels.append(Label(name=LabelType.FEATURE, value=feature.name))
        scenario_result.parameters = self.get_params(request.node)

    @pytest.hookimpl
    def pytest_bdd_after_scenario(self, request, feature, scenario):
        scenario_result_uuid = self._cache.get(scenario)
        scenario_result = self.allure_logger.get_item(scenario_result_uuid)
        scenario_result.stop = now()
        self.allure_logger.stop_step(scenario_result_uuid)

    @pytest.hookimpl
    def pytest_bdd_step_func_lookup_error(self, request, feature, scenario, step, exception):
        scenario_result_uuid = self._cache.get(scenario)
        scenario_result = self.allure_logger.get_item(scenario_result_uuid)
        scenario_result.status = Status.BROKEN
        scenario_result.status_details = exception
        self.allure_logger.stop_step(scenario_result_uuid)

    @staticmethod
    def get_params(node):
        if hasattr(node, "callspec"):
            params = node.callspec.params
            return [Parameter(name=name, value=value) for name, value in params.items()]

    @staticmethod
    def get_name(node, scenario):
        if hasattr(node, "callspec"):
            parts = node.nodeid.rsplit("[")
            return f"{scenario.name} [{parts[-1]}"
        return scenario.name

    @staticmethod
    def get_full_name(feature, scenario):
        feature_path = os.path.normpath(feature.rel_filename)
        return f"{feature_path}:{scenario.name}"



# Refactor to Enums
import re
from collections import defaultdict

TAG = "tag"
FEATURE = "feature"
SCENARIO_OUTLINE = "scenario outline"
EXAMPLES = "examples"
EXAMPLES_VERTICAL = "examples vertical"
SCENARIO = "scenario"
BACKGROUND = "background"
EXAMPLES_HEADERS = "example headers"
EXAMPLE_LINE = "example line"
EXAMPLE_LINE_VERTICAL = "example line vertical"
GIVEN = "given"
WHEN = "when"
THEN = "then"


class StepType:
    CONTEXT = "Context"
    ACTION = "Action"
    OUTCOME = "Outcome"
    CONJUNCTION = "Conjunction"
    UNSPECIFIED = "Unspecified"
    UNKNOWN = "Unknown"


STEP_PREFIXES = {
    FEATURE: "Feature: ",
    SCENARIO_OUTLINE: "Scenario Outline: ",
    EXAMPLES_VERTICAL: "Examples: Vertical",
    EXAMPLES: "Examples:",
    SCENARIO: "Scenario: ",
    BACKGROUND: "Background:",
    GIVEN: "Given ",
    WHEN: "When ",
    THEN: "Then ",
    TAG: "@",
}
STEP_TYPE_BY_NORMALIZED_PREFIX = defaultdict(
    lambda: StepType.UNKNOWN,
    {
        "given": StepType.CONTEXT,
        "when": StepType.ACTION,
        "then": StepType.OUTCOME,
        "and": StepType.CONJUNCTION,
        "but": StepType.CONJUNCTION,
        "*": StepType.CONJUNCTION,
    },
)
PYTHON_REPLACE_REGEX = re.compile(r"\W")
ALPHA_REGEX = re.compile(r"^\d+_*")

TYPE_KEYWORD_TYPE = defaultdict(
    lambda: StepType.UNKNOWN,
    {
        "And": StepType.CONJUNCTION,
        "But": StepType.CONJUNCTION,
        "*": StepType.UNSPECIFIED,
        "Given": StepType.CONTEXT,
        "When": StepType.ACTION,
        "Then": StepType.OUTCOME,
    },
)



"""StepHandler parsers."""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from re import Pattern as _RePattern
from re import compile as re_compile
from typing import Any, cast

import parse as base_parse
import parse_type.cfparse as base_cfparse

from pytest_bdd.typing import Protocol, runtime_checkable


@runtime_checkable
class StepParserProtocol(Protocol):
    name: Any

    def parse_arguments(self, name: str) -> dict[str, Any] | None:
        ...

    def is_matching(self, name: str) -> bool:
        ...


class StepParser(StepParserProtocol, metaclass=ABCMeta):
    """Parser of the individual step."""

    def __init__(self, name: Any) -> None:
        self.name = name

    @abstractmethod
    def parse_arguments(self, name: str) -> dict[str, Any] | None:
        """Get step arguments from the given step name.

        :return: `dict` of step arguments
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def is_matching(self, name: str) -> bool:
        """Match given name with the step name."""
        raise NotImplementedError()  # pragma: no cover


class _CommonRe(StepParser):
    regex: _RePattern

    def parse_arguments(self, name):
        return self.regex.match(name).groupdict()

    def is_matching(self, name):
        return bool(self.regex.match(name))


class _re(_CommonRe):
    def __init__(self, name):
        """Compile regex."""
        super().__init__(name.pattern)
        self.regex = name


class re(_CommonRe):
    """Regex step parser."""

    def __init__(self, name: str, *args: Any, **kwargs: Any) -> None:
        """Compile regex."""
        super().__init__(name)
        self.regex = re_compile(self.name, *args, **kwargs)


class _CommonParse(StepParser):
    parser: base_parse.Parser

    def parse_arguments(self, name: str) -> dict[str, Any] | None:
        return cast(dict, self.parser.parse(name).named)

    def is_matching(self, name):
        try:
            return bool(self.parser.parse(name))
        except ValueError:
            return False


class _parse(_CommonParse):
    def __init__(self, name: base_parse.Parser):
        super().__init__(name._format)
        self.parser = name


class parse(_CommonParse):
    """parse step parser."""

    def __init__(self, name: str, *args: Any, **kwargs: Any) -> None:
        """Compile parse expression."""
        super().__init__(name)
        self.parser = base_parse.compile(self.name, *args, **kwargs)


class cfparse(_CommonParse):
    """cfparse step parser."""

    def __init__(self, name, *args, **kwargs):
        """Stringify"""
        name = str(name, **({"encoding": "utf-8"} if isinstance(name, bytes) else {}))
        super().__init__(name)
        """Compile parse expression."""
        self.parser = base_cfparse.Parser(self.name, *args, **kwargs)


class string(StepParser):
    """Exact string step parser."""

    def __init__(self, name: str | bytes) -> None:
        """Stringify"""
        name = str(name, **({"encoding": "utf-8"} if isinstance(name, bytes) else {}))
        super().__init__(name)

    def parse_arguments(self, name: str) -> dict[str, Any]:
        """No parameters are available for simple string step.

        :return: `dict` of step arguments
        """
        return {}

    def is_matching(self, name: str) -> bool:
        """Match given name with the step name."""
        return bool(self.name == name)


def get_parser(parserlike: str | StepParser | StepParserProtocol) -> StepParser:
    """Get parser by given name.

    :param parserlike: name of the step to parse

    :return: step parser object
    :rtype: StepArgumentParser
    """

    if isinstance(parserlike, StepParserProtocol):
        return cast(StepParser, parserlike)
    elif isinstance(parserlike, _RePattern):
        return _re(parserlike)
    elif isinstance(parserlike, base_parse.Parser):
        return _parse(parserlike)
    else:
        return cfparse(parserlike)



"""StepHandler decorators.

Example:

@given("I have an article", target_fixture="article")
def given_article(author):
    return create_test_article(author=author)


@when("I go to the article page")
def go_to_the_article_page(browser, article):
    browser.visit(urljoin(browser.url, "/articles/{0}/".format(article.id)))


@then("I should not see the error message")
def no_error_message(browser):
    with pytest.raises(ElementDoesNotExist):
        browser.find_by_css(".message.error").first


Multiple names for the steps:

@given("I have an article")
@given("there is an article")
def article(author):
    return create_test_article(author=author)


Reusing existing fixtures for a different step name:


@given("I have a beautiful article")
def given_beautiful_article(article):
    pass

"""
from __future__ import annotations

import warnings
from contextlib import suppress
from typing import Any, Callable, Iterable, Iterator, Sequence
from warnings import warn

import pytest
from attr import Factory, attrib, attrs
from ordered_set import OrderedSet

from pytest_bdd.const import StepType
from pytest_bdd.model import Feature, Scenario, Step
from pytest_bdd.parsers import StepParser, get_parser
from pytest_bdd.typing.pytest import Config, Parser, TypeAlias
from pytest_bdd.utils import get_caller_module_locals
from pytest_bdd.warning_types import PytestBDDStepDefinitionWarning


def add_options(parser: Parser):
    """Add pytest-bdd options."""
    group = parser.getgroup("bdd", "Steps")
    group.addoption(
        "--liberal-steps",
        action="store_true",
        dest="liberal_steps",
        default=None,
        help="Allow use different keywords with same step definition",
    )
    parser.addini(
        "liberal_steps",
        default=False,
        type="bool",
        help="Allow use different keywords with same step definition",
    )


def given(
    parserlike: Any,
    converters: dict[str, Callable] | None = None,
    target_fixture: str | None = None,
    target_fixtures: list[str] = None,
    params_fixtures_mapping: set[str] | dict[str, str] | Any = True,
    param_defaults: dict | None = None,
    liberal: bool | None = None,
) -> Callable:
    """Given step decorator.

    :param parserlike: StepHandler name or a parser object.
    :param converters: Optional `dict` of the argument or parameter converters in form
                       {<param_name>: <converter function>}.
    :param target_fixture: Target fixture name to replace by steps definition function.
    :param target_fixtures: Target fixture names to be replaced by steps definition function.
    :param params_fixtures_mapping: StepHandler parameters would be injected as fixtures
    :param param_defaults: Default parameters for step definition
    :param liberal: Could step definition be used with other keywords

    :return: Decorator function for the step.
    """
    return StepHandler.decorator_builder(
        StepType.CONTEXT,
        parserlike,
        converters=converters,
        target_fixture=target_fixture,
        target_fixtures=target_fixtures,
        params_fixtures_mapping=params_fixtures_mapping,
        param_defaults=param_defaults,
        liberal=liberal,
    )


def when(
    parserlike: Any,
    converters: dict[str, Callable] | None = None,
    target_fixture: str | None = None,
    target_fixtures: list[str] = None,
    params_fixtures_mapping: set[str] | dict[str, str] | Any = True,
    param_defaults: dict | None = None,
    liberal: bool | None = None,
) -> Callable:
    """When step decorator.

    :param parserlike: StepHandler name or a parser object.
    :param converters: Optional `dict` of the argument or parameter converters in form
                       {<param_name>: <converter function>}.
    :param target_fixture: Target fixture name to replace by steps definition function.
    :param target_fixtures: Target fixture names to be replaced by steps definition function.
    :param params_fixtures_mapping: StepHandler parameters would be injected as fixtures
    :param param_defaults: Default parameters for step definition
    :param liberal: Could step definition be used with other keywords

    :return: Decorator function for the step.
    """
    return StepHandler.decorator_builder(
        StepType.ACTION,
        parserlike,
        converters=converters,
        target_fixture=target_fixture,
        target_fixtures=target_fixtures,
        params_fixtures_mapping=params_fixtures_mapping,
        param_defaults=param_defaults,
        liberal=liberal,
    )


def then(
    parserlike: Any,
    converters: dict[str, Callable] | None = None,
    target_fixture: str | None = None,
    target_fixtures: list[str] = None,
    params_fixtures_mapping: set[str] | dict[str, str] | Any = True,
    param_defaults: dict | None = None,
    liberal: bool | None = None,
) -> Callable:
    """Then step decorator.

    :param parserlike: StepHandler name or a parser object.
    :param converters: Optional `dict` of the argument or parameter converters in form
                       {<param_name>: <converter function>}.
    :param target_fixture: Target fixture name to replace by steps definition function.
    :param target_fixtures: Target fixture names to be replaced by steps definition function.
    :param params_fixtures_mapping: StepHandler parameters would be injected as fixtures
    :param param_defaults: Default parameters for step definition
    :param liberal: Could step definition be used with other keywords

    :return: Decorator function for the step.
    """
    return StepHandler.decorator_builder(
        StepType.OUTCOME,
        parserlike,
        converters=converters,
        target_fixture=target_fixture,
        target_fixtures=target_fixtures,
        params_fixtures_mapping=params_fixtures_mapping,
        param_defaults=param_defaults,
        liberal=liberal,
    )


def step(
    parserlike: Any,
    converters: dict[str, Callable] | None = None,
    target_fixture: str | None = None,
    target_fixtures: list[str] = None,
    params_fixtures_mapping: set[str] | dict[str, str] | Any = True,
    param_defaults: dict | None = None,
    liberal: bool | None = None,
):
    """Liberal step decorator which could be used with any keyword.

    :param parserlike: StepHandler name or a parser object.
    :param converters: Optional `dict` of the argument or parameter converters in form
                       {<param_name>: <converter function>}.
    :param target_fixture: Target fixture name to replace by steps definition function.
    :param target_fixtures: Target fixture names to be replaced by steps definition function.
    :param params_fixtures_mapping: StepHandler parameters would be injected as fixtures
    :param param_defaults: Default parameters for step definition
    :param liberal: Could step definition be used with other keywords

    :return: Decorator function for the step.
    """
    return StepHandler.decorator_builder(
        StepType.UNSPECIFIED if liberal else StepType.UNKNOWN,
        parserlike,
        converters=converters,
        target_fixture=target_fixture,
        target_fixtures=target_fixtures,
        params_fixtures_mapping=params_fixtures_mapping,
        param_defaults=param_defaults,
        liberal=liberal,
    )


class StepHandler:
    Model: TypeAlias = "Step"

    @attrs
    class Matcher:
        config: Config = attrib()
        feature: Feature = attrib(init=False)
        pickle: Scenario = attrib(init=False)
        step: Step = attrib(init=False)
        previous_step: Step | None = attrib(init=False)
        step_registry: StepHandler.Registry = attrib(init=False)
        step_type_context = attrib(default=None)

        class MatchNotFoundError(RuntimeError):
            pass

        def __call__(
            self,
            feature: Feature,
            pickle: Scenario,
            step: Step,
            previous_step: Step | None,
            step_registry: StepHandler.Registry,
        ) -> StepHandler.Definition:
            self.feature = feature
            self.pickle = pickle
            self.step = step
            self.previous_step = previous_step
            self.step_registry = step_registry

            self.step_type_context = (
                self.step_type_context
                if self.step.type in (StepType.CONJUNCTION, StepType.UNKNOWN) and self.step_type_context is not None
                else self.step.type
            )
            if self.step_type_context == StepType.CONJUNCTION:
                self.step_type_context = StepType.UNKNOWN

            step_definitions = list(
                self.find_step_definition_matches(
                    self.step_registry, (self.strict_matcher, self.unspecified_matcher, self.liberal_matcher)
                )
            )

            if len(step_definitions) > 0:
                if len(step_definitions) > 1:
                    warn(PytestBDDStepDefinitionWarning(f"Alternative step definitions are found: {step_definitions}"))
                return step_definitions[0]
            raise self.MatchNotFoundError

        def strict_matcher(self, step_definition):
            return step_definition.type_ == self.step_type_context and step_definition.parser.is_matching(
                self.step.text
            )

        def unspecified_matcher(self, step_definition):
            return (
                self.step_type_context == StepType.UNSPECIFIED or step_definition.type_ == StepType.UNSPECIFIED
            ) and step_definition.parser.is_matching(self.step.text)

        def liberal_matcher(self, step_definition):
            if step_definition.liberal is None:
                if self.config.option.liberal_steps is not None:
                    is_step_definition_liberal = self.config.option.liberal_steps
                else:
                    is_step_definition_liberal = self.config.getini("liberal_steps")
            else:
                is_step_definition_liberal = step_definition.liberal

            return all(
                (
                    not self.unspecified_matcher(step_definition),
                    is_step_definition_liberal,
                    step_definition.type_ != self.step_type_context,
                    step_definition.parser.is_matching(self.step.text),
                )
            )

        @staticmethod
        def find_step_definition_matches(
            registry: StepHandler.Registry | None, matchers: Sequence[Callable[[StepHandler.Definition], bool]]
        ) -> Iterable[StepHandler.Definition]:
            if registry:
                found_matches = False
                for matcher in matchers:
                    for step_definition in registry:
                        if matcher(step_definition):
                            found_matches = True
                            yield step_definition
                    if found_matches:
                        break
                if not found_matches:
                    with suppress(AttributeError):
                        yield from StepHandler.Matcher.find_step_definition_matches(registry.parent, matchers)

    @attrs(auto_attribs=True)
    class Definition:
        func: Callable
        type_: str | None
        parser: StepParser
        converters: dict[str, Any]
        params_fixtures_mapping: dict[str, str]
        param_defaults: dict
        target_fixtures: list[str]
        liberal: bool

        def get_parameters(self, step: Step):
            parsed_arguments = self.parser.parse_arguments(step.name) or {}
            return {
                **self.param_defaults,
                **{arg: self.converters.get(arg, lambda _: _)(value) for arg, value in parsed_arguments.items()},
            }

    @attrs
    class Registry:
        registry: list[StepHandler.Definition] = attrib(default=Factory(list))
        parent: StepHandler.Registry = attrib(default=None, init=False)

        @classmethod
        def register_step(
            cls,
            caller_locals: dict,
            func,
            type_,
            parserlike,
            converters,
            params_fixtures_mapping,
            param_defaults,
            target_fixtures,
            liberal,
        ):
            if "step_registry" not in caller_locals.keys():
                built_registry = cls()
                caller_locals["step_registry"] = built_registry.bind_pytest_bdd_step_registry_fixture()

            registry: StepHandler.Registry = caller_locals["step_registry"].__registry__

            parser = get_parser(parserlike)
            registry.registry.append(
                StepHandler.Definition(  # type: ignore[call-arg]
                    func=func,
                    type_=type_,
                    parser=parser,
                    converters=converters,
                    params_fixtures_mapping=params_fixtures_mapping,
                    param_defaults=param_defaults,
                    target_fixtures=target_fixtures,
                    liberal=liberal,
                )
            )

        def bind_pytest_bdd_step_registry_fixture(self):
            @pytest.fixture
            def step_registry(step_registry):
                self.parent = step_registry
                return self

            step_registry.__registry__ = self
            return step_registry

        def __iter__(self) -> Iterator[StepHandler.Definition]:
            return iter(self.registry)

    @staticmethod
    def decorator_builder(
        step_type: str | None,
        step_parserlike: Any,
        converters: dict[str, Callable] | None = None,
        target_fixture: str | None = None,
        target_fixtures: list[str] = None,
        params_fixtures_mapping: set[str] | dict[str, str] | Any = True,
        param_defaults: dict | None = None,
        liberal: bool | None = None,
    ) -> Callable:
        """StepHandler decorator for the type and the name.

        :param step_type: StepHandler type (CONTEXT, ACTION or OUTCOME).
        :param step_parserlike: StepHandler name as in the feature file.
        :param converters: Optional step arguments converters mapping
        :param target_fixture: Optional fixture name to replace by step definition
        :param target_fixtures: Target fixture names to be replaced by steps definition function.
        :param params_fixtures_mapping: StepHandler parameters would be injected as fixtures
        :param param_defaults: Default parameters for step definition
        :param liberal: Could step definition be used with other keywords

        :return: Decorator function for the step.
        """

        converters = converters or {}
        param_defaults = param_defaults or {}
        if target_fixture is not None and target_fixtures is not None:
            warnings.warn(PytestBDDStepDefinitionWarning("Both target_fixture and target_fixtures are specified"))
        target_fixtures = list(
            OrderedSet(
                [
                    *([target_fixture] if target_fixture is not None else []),
                    *(target_fixtures if target_fixtures is not None else []),
                ]
            )
        )

        def decorator(step_func: Callable) -> Callable:
            """
            StepHandler decorator

            :param function step_func: StepHandler definition function
            """
            StepHandler.Registry.register_step(
                caller_locals=get_caller_module_locals(depth=2),
                func=step_func,
                type_=step_type,
                parserlike=step_parserlike,
                converters=converters,
                params_fixtures_mapping=params_fixtures_mapping,
                param_defaults=param_defaults,
                target_fixtures=target_fixtures,
                liberal=liberal,
            )
            return step_func

        return decorator



"""Scenario implementation.

The pytest will collect the test case and the steps will be executed
line by line.

Example:

test_publish_article = scenario(
    feature_name="publish_article.feature",
    scenario_name="Publishing the article",
)
"""
from __future__ import annotations

import collections
from functools import reduce
from operator import truediv
from os.path import commonpath, dirname
from pathlib import Path
from typing import Callable, Iterable, Sequence, cast

import pytest
from attr import Factory, attrib, attrs

from pytest_bdd.model import Feature, Scenario
from pytest_bdd.parser import GherkinParser
from pytest_bdd.typing.parser import ParserProtocol
from pytest_bdd.typing.pytest import Config, Metafunc
from pytest_bdd.utils import get_caller_module_locals, get_caller_module_path, make_python_name

Args = collections.namedtuple("Args", ["args", "kwargs"])
FakeRequest = collections.namedtuple("FakeRequest", ["module"])


@attrs
class FileScenarioLocator:
    feature_paths = attrib(default=Factory(list))
    scenario_filter = attrib(default=None)
    encoding = attrib(default="utf-8")
    features_base_dir = attrib(default=None)
    parser: ParserProtocol = attrib(default=Factory(GherkinParser))
    parse_args: Args = attrib(default=Factory(lambda: Args((), {})))

    @property
    def glob(self):
        return self.parser.glob

    def resolve(self):
        features_base_dir = (
            self.features_base_dir() if isinstance(self.features_base_dir, Callable) else self.features_base_dir
        )
        features_base_dir = (Path.cwd() / Path(features_base_dir)).resolve()

        already_resolved = set()

        def feature_paths_gen():
            for feature_path in map(Path, self.feature_paths):
                if feature_path.is_dir():
                    yield from iter(self.glob(feature_path))
                else:
                    yield feature_path

        for feature_path in feature_paths_gen():
            if feature_path.is_absolute():
                try:
                    common_path = Path(commonpath([feature_path, features_base_dir]))
                except ValueError:
                    rel_feature_path = feature_path
                else:
                    sub_levels = len(features_base_dir.relative_to(common_path).parts)
                    sub_path = reduce(truediv, [".."] * sub_levels, Path())
                    rel_feature_path = sub_path / feature_path.relative_to(common_path)
            else:
                rel_feature_path = feature_path

            uri = str(rel_feature_path.as_posix())
            absolute_feature_path = (Path(features_base_dir) / rel_feature_path).resolve()

            if absolute_feature_path not in already_resolved:
                already_resolved.add(absolute_feature_path)

                feature = self.parser.parse(
                    absolute_feature_path,
                    uri,
                    *self.parse_args.args,
                    **{**dict(encoding=self.encoding), **self.parse_args.kwargs},
                )

                for scenario in feature.scenarios:
                    if self.scenario_filter is None or self.scenario_filter(feature, scenario):
                        yield feature, scenario


@attrs
class ModuleScenarioRegistry:

    caller_locals = attrib()
    caller_module_path = attrib()
    locator_registry: list[tuple] = attrib(default=Factory(list))
    resolved_locator_registry: dict = attrib(default=Factory(dict))
    resolved = attrib(default=False)
    config = attrib(default=None)

    @classmethod
    def get(
        cls,
        caller_locals: dict | None = None,
        caller_module_path: str = None,
    ) -> ModuleScenarioRegistry:
        caller_locals = cast(dict, caller_locals if caller_locals is not None else get_caller_module_locals(depth=2))
        caller_module_path = cast(
            str, caller_module_path if caller_module_path is not None else get_caller_module_path(depth=2)
        )
        if "__pytest_bdd_scenario_registry__" not in caller_locals.keys():
            caller_locals["__pytest_bdd_scenario_registry__"] = ModuleScenarioRegistry(  # type: ignore[call-arg]
                caller_locals=caller_locals, caller_module_path=caller_module_path
            )
        return cast(ModuleScenarioRegistry, caller_locals["__pytest_bdd_scenario_registry__"])

    def add(self, locator, test_func):
        test_func.__pytest_bdd_scenario_registry__ = self
        self.locator_registry.append((locator, test_func))
        if test_func.__name__ not in self.caller_locals.keys():
            self.caller_locals[test_func.__name__] = test_func

    def update(self, locator, updatable, test_func):
        for index, (registry_locator, registry_test_func) in enumerate(self.locator_registry):
            if locator == registry_locator:
                del self.caller_locals[updatable.__name__]
                self.add(locator, test_func)
                break

    @property
    def resolved_locators(self):
        if not self.resolved:
            self.resolved = True
            for locator, test_func in self.locator_registry:
                for feature, scenario in locator.resolve():
                    self.resolved_locator_registry[(feature.uri, scenario.id)] = test_func, feature, scenario
        return self.resolved_locator_registry

    def parametrize(self, metafunc: Metafunc):
        test_func = metafunc.function
        self.config = metafunc.config

        parametrizations = [
            (feature, scenario) for _, (func, feature, scenario) in self.resolved_locators.items() if func is test_func
        ]

        metafunc.parametrize(
            "feature, scenario",
            [self._build_scenario_param(feature, scenario, self.config) for feature, scenario in parametrizations],
        )

    @staticmethod
    def _build_scenario_param(feature: Feature, scenario: Scenario, config: Config):
        marks = []
        for tag in scenario.tag_names:
            tag_marks = config.hook.pytest_bdd_convert_tag_to_marks(feature=feature, scenario=scenario, tag=tag)
            if tag_marks is not None:
                marks.extend(tag_marks)
        return pytest.param(
            feature,
            scenario,
            id=f"{feature.uri}-{feature.name}-{scenario.name}{scenario.table_rows_breadcrumb}",
            marks=marks,
        )

    @property
    def features_base_dir(self) -> Path:
        default_base_dir = dirname(self.caller_module_path)
        return Path(self.get_from_config_ini("bdd_features_base_dir", default_base_dir))

    def get_from_config_ini(self, key: str, default: str) -> str:
        """Get value from ini config. Return default if value has not been set.

        Use if the default value is dynamic. Otherwise set default on addini call.
        """
        value = self.config.getini(key)
        if not isinstance(value, str):
            raise TypeError(f"Expected a string for configuration option {value!r}, got a {type(value)} instead")
        return value if value != "" else default

    def build_bound_locator_test_function(self, locator):
        @pytest.mark.pytest_bdd_scenario
        @pytest.mark.usefixtures("feature", "scenario")
        def test():
            ...

        test.__name__ = next(iter(test_names))

        self.add(locator, test)

        return test

    def build_bound_locator_test_decorator(self, locator):
        def decorator(func):
            updated_func = pytest.mark.pytest_bdd_scenario(pytest.mark.usefixtures("feature", "scenario")(func))
            self.update(locator, self.build_bound_locator_test_function(locator), updated_func)
            return updated_func

        return decorator


def get_python_name_generator(name: str) -> Iterable[str]:
    """Generate a sequence of suitable python names out of given arbitrary string name."""
    python_name = make_python_name(name)
    suffix = ""
    index = 0

    def get_name() -> str:
        return "_".join(filter(bool, ["test", python_name, suffix]))

    while True:
        yield get_name()
        index += 1
        suffix = f"{index}"


test_names = get_python_name_generator("")


def scenario(
    feature_name: Path | str = Path(),
    scenario_name: str | None = None,
    encoding: str = "utf-8",
    features_base_dir: Path | str | None = None,
    return_test_decorator=True,
    parser: ParserProtocol | None = None,
    parse_args=Args((), {}),
    _caller_module_locals=None,
    _caller_module_path=None,
):
    """
    Scenario decorator.

    :param feature_name: Feature file name. Absolute or relative to the configured feature base path.
    :param scenario_name: Scenario name.
    :param encoding: Feature file encoding.
    :param features_base_dir: Feature base directory from where features will be searched
    :param return_test_decorator; Return test decorator or generated test
    :param parser: Parser used to parse feature-like file
    :param parse_args: args consumed by parser during parsing
    """
    return _scenarios(
        feature_paths=[feature_name],
        scenario_filter_or_scenario_name=scenario_name,
        encoding=encoding,
        features_base_dir=features_base_dir,
        return_test_decorator=return_test_decorator,
        _caller_module_locals=_caller_module_locals or get_caller_module_locals(depth=2),
        _caller_module_path=_caller_module_path or get_caller_module_path(depth=2),
        parser=parser,
        parse_args=parse_args,
    )


def scenarios(
    *feature_paths: Path | str,
    encoding: str = "utf-8",
    features_base_dir: Path | str | None = None,
    return_test_decorator=False,
    parser: ParserProtocol | None = None,
    parse_args=Args((), {}),
    _caller_module_locals=None,
    _caller_module_path=None,
):
    """
    Scenario decorator.

    :param feature_paths: Features file names. Absolute or relative to the configured feature base path.
    :param encoding: Feature file encoding.
    :param features_base_dir: Feature base directory from where features will be searched
    :param return_test_decorator; Return test decorator or generated test
    """
    return _scenarios(
        feature_paths=feature_paths,
        scenario_filter_or_scenario_name=None,
        encoding=encoding,
        features_base_dir=features_base_dir,
        return_test_decorator=return_test_decorator,
        parser=parser,
        parse_args=parse_args,
        _caller_module_locals=_caller_module_locals or get_caller_module_locals(depth=2),
        _caller_module_path=_caller_module_path or get_caller_module_path(depth=2),
    )


def _scenarios(
    feature_paths: Sequence[Path | str],
    scenario_filter_or_scenario_name: str | Callable | None,
    return_test_decorator=True,
    encoding: str = "utf-8",
    features_base_dir: Path | str | None = None,
    parser: ParserProtocol | None = None,
    parse_args=Args((), {}),
    _caller_module_locals=None,
    _caller_module_path=None,
):
    if parser is None:
        parser = GherkinParser()
    module_scenario_registry = ModuleScenarioRegistry.get(
        caller_locals=_caller_module_locals,
        caller_module_path=_caller_module_path,
    )

    scenario_filter_kwarg = {}
    if isinstance(scenario_filter_or_scenario_name, str):
        scenario_filter_kwarg = {
            "scenario_filter": lambda feature, scenario: scenario.name == scenario_filter_or_scenario_name
        }
    if callable(scenario_filter_or_scenario_name):
        scenario_filter_kwarg = {"scenario_filter": scenario_filter_or_scenario_name}

    scenario_locator = FileScenarioLocator(  # type: ignore[call-arg]
        feature_paths=feature_paths,
        **scenario_filter_kwarg,
        encoding=encoding,
        features_base_dir=(
            features_base_dir if features_base_dir is not None else lambda: module_scenario_registry.features_base_dir
        ),  # known only after start of pytest runtime
        parser=parser,
        parse_args=parse_args,
    )

    return (
        module_scenario_registry.build_bound_locator_test_decorator
        if return_test_decorator
        else module_scenario_registry.build_bound_locator_test_function
    )(scenario_locator)



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, cast

from attr import Factory, attrib, attrs
from marshmallow import Schema, fields, post_load

from pytest_bdd.ast import Scenario as ASTScenario
from pytest_bdd.ast import Step as ASTStep
from pytest_bdd.ast import TableRow as ASTTableRow
from pytest_bdd.const import STEP_PREFIXES, TAG, TYPE_KEYWORD_TYPE
from pytest_bdd.utils import ModelSchemaPostLoadable, _itemgetter, deepattrgetter

if TYPE_CHECKING:  # pragma: no cover
    from pytest_bdd.model.feature import Feature


@attrs
class ASTLinkMixin:
    ast = attrib(init=False)

    @property
    def linked_ast_nodes(self):
        return _itemgetter(
            *((self.ast_node_id,) if hasattr(self, "ast_node_id") else ()),
            *getattr(self, "ast_node_ids", ()),
        )(self.ast.registry)


@attrs
class ASTNodeIDsMixin(ASTLinkMixin):
    ast_node_ids: list[str] = attrib()


class ASTNodeIDsSchemaMixin(Schema):
    ast_node_ids = fields.List(fields.Str(), data_key="astNodeIds")


@attrs
class Cell(ModelSchemaPostLoadable):
    value: str = attrib()


class CellSchema(Schema):
    value = fields.Str()

    build_cell = Cell.schema_post_loader()


@attrs
class Row(ModelSchemaPostLoadable):
    cells: list[Cell] = attrib()


class RowSchema(Schema):
    cells = fields.Nested(CellSchema(many=True))

    build_row = Row.schema_post_loader()


@attrs
class DataTable(ModelSchemaPostLoadable):
    rows: list[Row] = attrib()


class DataTableSchema(Schema):
    rows = fields.Nested(RowSchema(many=True))

    build_data_table = DataTable.schema_post_loader()


@attrs
class DocString(ModelSchemaPostLoadable):
    content: str = attrib()
    # Workaround because of allure integration
    if TYPE_CHECKING:  # pragma: no cover
        media_type: str = attrib(init=False)

    postbuild_attrs = ["media_type"]


class DocStringSchema(Schema):
    content = fields.Str()
    media_type = fields.Str(data_key="mediaType", required=False)

    build_doc_string = DocString.schema_post_loader()


@attrs
class Argument(ModelSchemaPostLoadable):
    # Workaround because of allure integration
    if TYPE_CHECKING:  # pragma: no cover
        data_table: DataTable = attrib(init=False)
        doc_string: DocString = attrib(init=False)

    postbuild_attrs = ["data_table", "doc_string"]


class ArgumentSchema(Schema):
    data_table = fields.Nested(DataTableSchema(), data_key="dataTable", required=False)
    doc_string = fields.Nested(DocStringSchema(), data_key="docString", required=False)

    build_argument = Argument.schema_post_loader()


@attrs
class Step(ASTNodeIDsMixin, ModelSchemaPostLoadable):
    id: str = attrib()
    text: str = attrib()
    type: str = attrib()
    # Workaround because of allure integration
    if TYPE_CHECKING:  # pragma: no cover
        argument: Argument = attrib(init=False)

    postbuild_attrs = ["argument"]

    # region Indirectly loadable
    scenario = attrib(init=False)
    # endregion

    @property
    def _ast_step(self) -> ASTStep:
        return cast(ASTStep, next(filter(lambda node: type(node) is ASTStep, self.linked_ast_nodes)))

    @property
    def prefix(self):
        return self.keyword.lower()

    @property
    def name(self):
        return self.text

    @property
    def keyword(self):
        return self._ast_step.keyword.strip()

    @property
    def line_number(self):
        return self._ast_step.location.line

    @property
    def doc_string(self):
        return self._ast_step.doc_string

    @property
    def data_table(self):
        return self._ast_step.data_table

    def decompose(self):
        self.ast = None
        self.scenario = None


class StepSchema(ASTNodeIDsSchemaMixin, Schema):
    argument = fields.Nested(ArgumentSchema(), required=False)
    id = fields.Str()
    text = fields.Str()
    type = fields.Str()

    build_step = Step.schema_post_loader()


@attrs
class Tag(ASTLinkMixin, ModelSchemaPostLoadable):
    ast_node_id: str = attrib()
    name: str = attrib()

    def decompose(self):
        self.ast = None


class TagSchema(Schema):
    ast_node_id = fields.Str(data_key="astNodeId")
    name = fields.Str()

    build_tag = Tag.schema_post_loader()


@attrs
class Scenario(ASTNodeIDsMixin, ModelSchemaPostLoadable):
    id: str = attrib()
    name: str = attrib()
    language: str = attrib()
    steps: list[Step] = attrib()
    tags: list[Tag] = attrib()
    uri: str = attrib()

    # region Indirectly loadable
    feature: Feature = attrib(init=False)
    # endregion

    @property
    def _ast_scenario(self) -> ASTScenario:
        return cast(ASTScenario, next(filter(lambda node: type(node) is ASTScenario, self.linked_ast_nodes)))

    @property
    def _ast_table_rows(self) -> list[ASTTableRow]:
        return list(filter(lambda node: type(node) is ASTTableRow, self.linked_ast_nodes))

    @property
    def table_rows_breadcrumb(self):
        table_rows_lines = ",".join(
            map(lambda row: f"line: {deepattrgetter('location.line', default=-1)(row)[0]}", self._ast_table_rows)
        )
        return f"[table_rows:[{table_rows_lines}]]" if table_rows_lines else ""

    @property
    def line_number(self):
        return self._ast_scenario.location.line

    @property
    def tag_names(self):
        return sorted(map(lambda tag: tag.name.lstrip(STEP_PREFIXES[TAG]), self.tags))

    def bind_ast(self, ast):
        self.ast = ast

        for step in self.steps:
            step.ast = ast

        for tag in self.tags:
            tag.ast = ast

    def bind_feature(self, feature: Feature):
        self.feature = feature

        self.bind_ast(feature.gherkin_ast)

    def bind_steps(self):
        for step in self.steps:
            step.scenario = self

    def decompose(self):
        self.feature = None
        self.ast = None
        for step in self.steps:
            step.decompose()
        for tag in self.tags:
            tag.decompose()


class ScenarioSchema(ASTNodeIDsSchemaMixin, Schema):
    id = fields.Str()
    name = fields.Str()
    language = fields.Str()
    steps = fields.Nested(StepSchema(many=True))
    tags = fields.Nested(TagSchema(many=True))
    uri = fields.Str()

    @post_load
    def build_scenario(self, data, many, **kwargs):
        scenario = Scenario(**data)
        scenario.bind_steps()
        return scenario


@attrs
class UserStep:
    text: str = attrib()
    scenario = attrib()

    id: str = attrib(default=Factory(lambda: str(uuid.uuid4())))
    argument: Argument = attrib(default=None)
    keyword = attrib(default="Given")
    line_number = attrib(default=-1)
    doc_string: DocString | None = attrib(default=None)
    data_table: DataTable | None = attrib(default=None)

    @property
    def prefix(self):
        return self.keyword.lower()

    @property
    def name(self):
        return self.text

    def decompose(self):
        self.scenario = None

    @property
    def type(self):
        return TYPE_KEYWORD_TYPE[self.keyword]



"""Reporting functionality.

Collection of the scenario execution statuses, timing and other information
that enriches the pytest test reporting.
"""
from __future__ import annotations

import time
from typing import Any, Callable

import pytest
from attr import Factory, attrib, attrs

from pytest_bdd.model import Feature, Scenario, Step
from pytest_bdd.typing.pytest import CallInfo, FixtureRequest, Item


class StepReport:
    """StepHandler execution report."""

    failed = False
    stopped = None

    def __init__(self, step: Step) -> None:
        """StepHandler report constructor.

        :param StepHandler step: StepHandler.
        """
        self.step = step
        self.started = time.perf_counter()

    def serialize(self) -> dict[str, Any]:
        """Serialize the step execution report.

        :return: Serialized step execution report.
        :rtype: dict
        """
        return {
            "name": self.step.name,
            "type": self.step.prefix,
            "keyword": self.step.keyword,
            "line_number": self.step.line_number,
            "failed": self.failed,
            "duration": self.duration,
        }

    def finalize(self, failed: bool) -> None:
        """Stop collecting information and finalize the report.

        :param bool failed: Whether the step execution is failed.
        """
        self.stopped = time.perf_counter()
        self.failed = failed

    @property
    def duration(self) -> float:
        """StepHandler execution duration.

        :return: StepHandler execution duration.
        :rtype: float
        """
        if self.stopped is None:
            return 0

        return self.stopped - self.started


@attrs
class ScenarioReport:
    """Scenario execution report."""

    feature: Feature = attrib()
    scenario: Scenario = attrib()
    step_reports: list[StepReport] = attrib(default=Factory(list))

    @property
    def current_step_report(self) -> StepReport:
        """Get current step report.

        :return: Last or current step report.
        :rtype: pytest_bdd.reporting.StepReport
        """
        return self.step_reports[-1]

    def add_step_report(self, step_report: StepReport) -> None:
        """Add new step report.

        :param step_report: New current step report.
        :type step_report: pytest_bdd.reporting.StepReport
        """
        self.step_reports.append(step_report)

    def serialize(self) -> dict[str, Any]:
        """Serialize scenario execution report in order to transfer reporting from nodes in the distributed mode.

        :return: Serialized report.
        :rtype: dict
        """
        pickle = self.scenario
        feature = self.feature

        return {
            "steps": [step_report.serialize() for step_report in self.step_reports],
            "name": pickle.name,
            "line_number": pickle.line_number,
            "tags": sorted(set(pickle.tag_names).difference(feature.tag_names)),
            "feature": {
                "name": feature.name,
                "filename": feature.filename,
                "rel_filename": feature.rel_filename,
                "line_number": feature.line_number,
                "description": feature.description,
                "tags": feature.tag_names,
            },
        }

    def fail(self) -> None:
        """Stop collecting information and finalize the report as failed."""
        self.current_step_report.finalize(failed=True)
        remaining_steps = self.scenario.steps[len(self.step_reports) :]

        # Fail the rest of the steps and make reports.
        for step in remaining_steps:
            report = StepReport(step=step)
            report.finalize(failed=True)
            self.add_step_report(report)


class ScenarioReporterPlugin:
    def __init__(self):
        self.current_report = None

    @pytest.mark.hookwrapper
    def pytest_runtest_makereport(self, item: Item, call: CallInfo):
        outcome = yield
        if call.when != "setup":
            rep = outcome.get_result()
            """Store item in the report object."""
            scenario_report: ScenarioReport = self.current_report

            if scenario_report is not None:
                rep.scenario = scenario_report.serialize()
                rep.item = {"name": item.name}

    @pytest.mark.tryfirst
    def pytest_bdd_before_scenario(self, request: FixtureRequest, feature: Feature, scenario: Scenario) -> None:
        """Create scenario report for the item."""
        self.current_report = ScenarioReport(feature=feature, scenario=scenario)  # type: ignore[call-arg]

    @pytest.mark.tryfirst
    def pytest_bdd_step_error(
        self,
        request: FixtureRequest,
        feature: Feature,
        scenario: Scenario,
        step: Step,
        step_func: Callable,
        step_func_args: dict,
        exception: Exception,
    ) -> None:
        """Finalize the step report as failed."""
        self.current_report.fail()

    @pytest.mark.tryfirst
    def pytest_bdd_before_step(
        self, request: FixtureRequest, feature: Feature, scenario: Scenario, step: Step, step_func: Callable
    ) -> None:
        """Store step start time."""
        self.current_report.add_step_report(StepReport(step=step))

    @pytest.mark.tryfirst
    def pytest_bdd_after_step(
        self,
        request: FixtureRequest,
        feature: Feature,
        scenario: Scenario,
        step: Step,
        step_func: Callable,
        step_func_args: dict,
    ) -> None:
        """Finalize the step report as successful."""
        self.current_report.current_step_report.finalize(failed=False)
